#include <Eigen/Dense>
#include <algorithm>
#include <array>
#include <bitset>
#include <execution>
#include <functional>
#include <iostream>
#include <map>
#include <memory>
#include <mutex>

template <unsigned int dim, typename number, typename T> class Orthtree {
public:
  typedef Eigen::Vector<number, dim> Point;
  typedef std::function<bool(std::unique_ptr<T> &, const Point &)>
      data_access_f;
  typedef std::function<bool(const std::unique_ptr<T> &, const Point &)>
      cdata_access_f;

  class BBox {
  public:
    enum Relation {
      NO_OVERLAP,
      OVERLAP,
      CONTAINS,
      CONTAINED,
    };
    Point min, max;
    Point center() { return (min + max) / 2; }
    Relation relation(const BBox &b) {
      if ((b.min.array() > max.array()).any() ||
          (b.max.array() < min.array()).any())
        return NO_OVERLAP;
      if ((b.min.array() > min.array()).all() &&
          (b.max.array() < max.array()).all())
        return CONTAINS;
      if ((b.min.array() < min.array()).all() &&
          (b.max.array() > max.array()).all())
        return CONTAINED;
      return OVERLAP;
    };
    BBox(const Point &min, const Point &max) : min(min), max(max) {
      assert((min.array() < max.array()).all());
    }
  };
  class Node {
  public:
    typedef std::bitset<dim> coord;
    Point center;
    std::unique_ptr<std::array<Node, (2 << dim)>> children;
    std::unique_ptr<T> data;

    mutable std::mutex mtx;

    void create_children(const Point &parent_center, number parent_width) {
      children = std::make_unique<std::array<Node, (2 << dim)>>();
      for (unsigned int i = 0; i < (2 << dim); ++i) {
        Point offset = Point::Ones() * (parent_width / 4);
        for (unsigned int j = 0; j < dim; ++j)
          if (!coord(i)[j])
            offset(j) *= -1;
        (*children)[i].center = parent_center + offset;
      }
    };

    unsigned int get_child_coord(const Point &p) const {
      coord child_coord;
      for (unsigned int i = 0; i < dim; ++i)
        child_coord[i] = center[i] < p[i];
      return child_coord.to_ulong();
    };

    void depth(const Point &point, int depth, data_access_f f, number width) {
      if (depth == 0) {
        std::lock_guard<std::mutex> guard(mtx);
        f(data, center);
      } else {
        {
          std::lock_guard<std::mutex> guard(mtx);
          if (!children)
            create_children(center, width);
        }
        (*children)[get_child_coord(point)].depth(point, depth - 1, f,
                                                  width / 2);
      }
    };
    BBox bb(number width) const {
      Point offset = Point::Ones() * (width / 2);
      return BBox(center - offset, center + offset);
    };
    void access_data(cdata_access_f f) const {
      std::lock_guard<std::mutex> guard(mtx);
      if (data) {
        f(data, center);
      }
    }
    void all_data(cdata_access_f f) const {
      std::lock_guard<std::mutex> guard(mtx);
      if (data) {
        f(data, center);
      }
      if (children) {
        std::for_each(std::execution::par_unseq, std::begin(*children),
                      std::end(*children),
                      [f](const auto &c) { c.all_data(f); });
      }
    }
    bool is_leaf() const {
      std::lock_guard<std::mutex> guard(mtx);
      return !bool(children);
    }
  };
  std::unique_ptr<Node> root;
  number root_width;

  mutable std::mutex mtx;
  Orthtree(const Point &root_center, number root_width)
      : root(new Node()), root_width(root_width) {
    root->center = root_center;
  };

  void all_data(cdata_access_f f) const { root->all_data(f); };

  void depth(const Point &point, int depth, data_access_f f) {
    root->depth(point, depth, f, root_width);
  }

  void data_in_bb(const BBox &bb, cdata_access_f f) const {
    number curr_width = root_width;
    std::vector<const Node *> to_visit{root.get()};
    std::vector<const Node *> to_visit_children;
    bool stop = false;
    while (!to_visit.empty() && !stop) {
      auto cbb = to_visit.back()->bb(curr_width);
      auto bbrel = cbb.relation(bb);
      // std::cout << to_visit.back()->center.transpose() << " - "
                // << cbb.min.transpose() << " - " << bbrel << std::endl;
      switch (bbrel) {
      case (Orthtree::BBox::CONTAINED):
        to_visit.back()->all_data(f);
        break;
      case (Orthtree::BBox::OVERLAP):
      case Orthtree::BBox::CONTAINS:
        to_visit.back()->access_data([&](auto &p, auto c) {
          if (!f(p, c))
            stop = true;
          return false;
        });
        if (!to_visit.back()->is_leaf()) {
          // std::transform(std::execution::par_unseq,to_visit.back()->children->cbegin(),
          //                to_visit.back()->children->cend(),
          //                to_visit_children.begin(), std::addressof<const
          //                Node>);
          for (const auto &n : *to_visit.back()->children) {
            to_visit_children.push_back(&n);
          }
          // std::cout << "leaf "<<to_visit_children.size() << std::endl;
        }
        break;
      default:
        break;
      }
      to_visit.pop_back();
      if (to_visit.empty()) {
        to_visit.swap(to_visit_children);
        to_visit_children.clear();
        curr_width /= 2;
      }
    }
  }
};
