#include <memory>

#include "nav_msgs/msg/map_meta_data.hpp"
#include "nav_msgs/msg/occupancy_grid.hpp"
#include "nav_msgs/srv/get_map.hpp"
#include "rclcpp/rclcpp.hpp"
#include "sensor_msgs/msg/laser_scan.hpp"

#include <Eigen/Dense>
#include <algorithm>
#include <limits>
#include <vector>

#include "simpleslam.hpp"
#include "utils.hpp"
#include "orthtree.hpp"

class MinimalSubscriber : public rclcpp::Node {
public:
  SimpleSlam slam;
  MinimalSubscriber() : Node("minimal_subscriber") {
    pub_map_ = this->create_publisher<nav_msgs::msg::OccupancyGrid>("/map", 10);
    auto topic_callback =
        [this](sensor_msgs::msg::LaserScan::UniquePtr msg) -> void {
      // RCLCPP_INFO(this->get_logger(), "angle min '%f'", msg->angle_min);
      // RCLCPP_INFO(this->get_logger(), "angle max '%f'", msg->angle_max);
      // RCLCPP_INFO(this->get_logger(), "time_increment'%f'",
      //             msg->time_increment);
      // RCLCPP_INFO(this->get_logger(), "angle_increment '%f'",
      //             msg->angle_increment);
      std::vector<Eigen::Vector2f> scan;

      Eigen::Vector2f min = Eigen::Vector2f(std::numeric_limits<float>::max(),
                                            std::numeric_limits<float>::max());
      Eigen::Vector2f max = Eigen::Vector2f(std::numeric_limits<float>::min(),
                                            std::numeric_limits<float>::min());
      for (size_t i = 0; i < msg->ranges.size(); i += 1) {
        auto radius = msg->ranges[i];
        Eigen::Rotation2D<float> rot2(msg->angle_min +
                                      i * msg->angle_increment);

        auto p = rot2 * Eigen::Vector2f(radius, 0);
        if (p.array().isInf().any())
          continue;
        min = min.cwiseMin(p);
        max = max.cwiseMax(p);
        scan.push_back(p);
        // RCLCPP_INFO(this->get_logger(), "p(%f,%f)",p[0],p[1]);
      }
      // std::transform(std::views::enumerate(msg->ranges).cbegin(),
      //                std::views::enumerate(msg->ranges).cend(), scan.begin(),
      //                [this](const auto &pair) {
      //                  const auto &[index, radius] = pair;
      //                  Eigen::Rotation2D<float> rot2(
      //                      msg->angle_min + index * msg->angle_increment);
      //                  return rot2 * Eigen::Vector2f(radius, 0);
      //                });

      nav_msgs::msg::OccupancyGrid grid;
      grid.header.frame_id = "map";
      grid.info.resolution = 50.0f / 1000.0f; /* [m/cell] */
      grid.info.width = (max - min)[0] / grid.info.resolution + 1;
      grid.info.height = (max - min)[1] / grid.info.resolution + 1;
      grid.info.origin.position.x = min[0];
      grid.info.origin.position.y = min[1];
      grid.data.resize(grid.info.width * grid.info.height);
      std::fill(grid.data.begin(), grid.data.end(), -1);
      auto put_p = [&](const Eigen::Vector2i &pi, const int o) {
        int idx = pi.x() + pi.y() * grid.info.width;
        // if(idx < grid.data.size()) {
        if (grid.data[idx] <= 0)
          grid.data[idx] = o;
      };
      for (const auto &v : scan) {
        Eigen::Vector2i vi = ((v - min) / grid.info.resolution).cast<int>();
        bresenham((-min / grid.info.resolution).cast<int>(), vi,
                  [&](const auto &p) { put_p(p, 0); });
        put_p(vi, 100);
      }
      pub_map_->publish(grid);
    };
    subscription_ = this->create_subscription<sensor_msgs::msg::LaserScan>(
        "scan", 10, topic_callback);
  }

private:
  rclcpp::Subscription<sensor_msgs::msg::LaserScan>::SharedPtr subscription_;
  rclcpp::Publisher<nav_msgs::msg::OccupancyGrid>::SharedPtr pub_map_;
};

int main(int argc, char *argv[]) {
  rclcpp::init(argc, argv);
  rclcpp::spin(std::make_shared<MinimalSubscriber>());
  rclcpp::shutdown();
  return 0;
}
